/* Example solution Oppgave 2, Oblig 2, INF4130, h09 
 *
 * Matias Holte (kjetimh) 
 *
 * if you have suggestions how to improve the code, you are welcome
 * to complain, like I complained on your code...
 */

import java.util.*;
import java.io.*;
import java.lang.*;

class Board{
    // store board as byte array
    byte[] board;
    // position of zero square
    int zero;

    public Board(byte[] b){
	board = new byte[b.length];
	for (int i=0;i<b.length;i++){
	    board[i]=b[i];
	    if (b[i] == 0)
		zero = i;
	}
    }
    public Board(Board b){
	board = new byte[b.board.length];
	for (int i=0;i<b.board.length;i++){
	    this.board[i] = b.board[i];
	}
	this.zero = b.zero;
    }
    // override equals: equal if board is equal
    public boolean equals(Object obj){
	if (this == obj)
	    return true;
	Board b2 = (Board) obj;
	for (int i=0;i<this.board.length;i++){
	    if (this.board[i] != b2.board[i])
		return false;
	}
	return true;
    }
    // override hashCode: calculate something from board
    public int hashCode(){
	int sum=0;
	for (int i=0;i<board.length;i++){
	    sum*=29;
	    sum+=board[i];
	}
	return sum;
    }
    //  manhattan heuristic
    int heuristic(){
	int sum = 0;
	int n = Oppgave2.N;
	for (int i=0;i<board.length;i++){
	    if (board[i]!=0){ // ignore 0 square
		// horizontal distance
		sum+=Math.abs(i%n - (board[i]-1)%n);
		// vertical distance
		sum+=Math.abs(i/n - (board[i]-1)/n);
	    }
	}
	return sum;
    }
};

class Node implements Comparable<Node>{
    // manhattan heuristic
    int heuristic;
    // distance from start
    int distance;
    // distance + heuristic
    int cost;
    Board board;
    // which node we came from
    Node parent;
    // which direction we used to get here
    char cparent;
    public Node(Board b,Node par,int h, int d, int c,char cpar){
	this.board = b;
	this.heuristic = h;
	this.distance = d;
	this.cost = c;
	this.parent = par;
	this.cparent = cpar;
    }
    // used by priority queue
    public int compareTo(Node a){
        if (a == null) return -1;
        if (this.cost<a.cost) return -1;
        if (this.cost>a.cost) return 1;
	return 0;
    }
    // Get one of the 4 neighboring boards
    Node getnumber(int i){
	if (i==0) return getleft();
	if (i==1) return getright();
	if (i==2) return getup();
	if (i==3) return getdown();
	return null;
    }
    Node getleft(){
	// if possible to go left
	if (this.board.zero % Oppgave2.N == 0)
	    return null;
	// go left (exchange zero with zero-1)
	return getNode(this.board.zero-1,'V');
    }
    Node getright(){
	if ((this.board.zero+1) % Oppgave2.N == 0)
	    return null;
	return getNode(this.board.zero+1,'H');
    }
    Node getup(){
	if (this.board.zero < Oppgave2.N)
	    return null;
	return getNode(this.board.zero-Oppgave2.N,'O');
    }
    Node getdown(){
	if (this.board.zero / Oppgave2.N >= Oppgave2.N-1)
	    return null;
	return getNode(this.board.zero+Oppgave2.N,'N');
    }

    // the function doing the actual work - create a new Node with a
    // Board where the zero field is moved to swapWith. 
    Node getNode(int swapWith,char parent){
	Board newboard = new Board(this.board);
	newboard.board[newboard.zero] = newboard.board[swapWith];
	newboard.board[swapWith] = 0;
	newboard.zero = swapWith;
	int newheu = newboard.heuristic();
	return new Node(newboard,this,newheu,this.distance+1,this.distance+1+newheu,parent);
    }
}

class Oppgave2{
    static int N;
    byte[] input;
    Scanner in;
    PrintWriter out;
    // HashMap over all visited boards (seen AND visited)
    HashMap<Board,Node> visitedBoards;
    // PriorityQueue over all seen Nodes, but not visited Nodes.
    PriorityQueue<Node> priQ;
    Node goalnode;
    Board goalboard;
    // number of states taken out of priQ
    int numtilstander=0;
    // number of states added to priQ
    int nytilstand=0;
    // number of duplicates visited
    int reinsert=0;
    public Oppgave2(String[] args){
	try{
	    in = new Scanner(new FileReader(args[0]));
	}
	catch(Exception e){
	    in = new Scanner(System.in);
	}
	try{
	    out = new PrintWriter(new File(args[1]));
	}
	catch(Exception e){
	    out = new PrintWriter(System.out);
	}
	// read input
	try{
	    N = in.nextInt();
	    input = new byte[N*N];
	    for (int i=0;i<N*N;i++){
		input[i] = (byte)in.nextInt();
	    }
	}
	catch(Exception e){
	    System.out.println(e);
	}
	Board startboard = new Board(input);

	// create goalboard
	for (int i=0;i<N*N-1;i++){
	    input[i] = (byte)(i+1);
	}
	input[N*N-1] = 0;
	goalboard = new Board(input);
	int starthau = startboard.heuristic();
	Node startnode = new Node(startboard,null,starthau,0,starthau,' ');
	priQ = new PriorityQueue<Node>();
	visitedBoards = new HashMap<Board,Node>();
	// add the first node so that we are able to start searching
	priQ.add(startnode);
	if(startsearch())
	    backtrack();
	in.close();
	out.close();
    }

    boolean startsearch(){
	while(true){
	    // take the best node out of the queue
	    Node cur = priQ.poll();
	    // no nodes left, didn't find a solution, return false
	    if (cur == null) return false;
	    numtilstander++;
	    // if it has been visited before, discard it
	    if (visitedBoards.get(cur.board) != null){
		reinsert++;
		continue;
	    }
	    // set it as visited
	    visitedBoards.put(cur.board,cur);
	    // if goalnode, store node in goalnode for backtracking,
	    // and return true (found)
	    if (cur.heuristic == 0 && cur.board.equals(goalboard)){
		goalnode = cur;
		return true;
	    }
	    // try to visit all neighbors by adding them to the
	    // queue. Possible optimization: don't visit parent node
	    // (again). 
	    for (int i=0;i<4;i++){
		Node next = cur.getnumber(i);
		if (next != null){
		    nytilstand++;
		    priQ.add(next);
		}
	    }
	}
    }
    void backtrack(){
	out.println(goalnode.cost);
	String res="";
	// follow parent pointers until start, and print out path
	while(goalnode.parent != null){
	    res = goalnode.cparent + res;
	    goalnode = goalnode.parent;
	}
	out.println(res);
	out.println(numtilstander);
	out.println(nytilstand);
	out.println(reinsert);
    }
    public static void main(String[] args){
	new Oppgave2(args);
    }
}
