# -*- coding: utf-8 -*-
import sys

class LeftistNode( object ):

    def __init__( self, element, left=None, right=None ):

        self.element = element   # prioriteten
        self.left    = left      # ventre barn
        self.right   = right     # høyre barn
        self.npl     = 0         # null distance lenght


    def printTree( self ):
        """
        Skriver ut treet infix. Brukes når man skriver til fil.
        """
        infix = []
        if self.left is not None:
            infix.extend( self.left.printTree() )
        infix.append(str(self.element))
        if self.right is not None:
            infix.extend( self.right.printTree() )
        return infix


    def printNullPath( self ):
        """
        Skriver ut stien fra rootnoden til den første noden som
        har nullsti lengde lik 0 Brukes når man skriver til fil.
        """
        infix = [str(self.element)]
        if self.npl == 0:
            pass
        elif self.left.npl <= self.right.npl:
            data = self.left.printNullPath()
            infix.extend(data)
        else:
            data = self.right.printNullPath()
            infix.extend(data)
        return infix

    

class LeftistHeap( object ):

    def __init__( self ):
        self.root   = None

    def deleteMin( self ):
        """
        Sletter elementet i heapen med lavest prioritet. Ikke gjort noen
        forandringer her i forhold til MAW implementasjonen
        """
        if self.root == None:
            return None
        min = self.root
        self.root = self.merge1( self.root.left, self.root.right )
        return min

    
    def insert( self, element ):
        """
        Setter inn et element i heapen. Elementet blir gjort om til en
        LeftistNode. Dette gjøres lett ved å sette rootnoden til å
        være resultatet av en merge1 mellem elementet og den allerede
        eksisterende rootnoden. Ikke noe forskjell her heller i forhold
        til MAW implementasjonen
        """
        self.root = self.merge1( LeftistNode(element), self.root )


    def merge( self, rhs ):
        """
        Merger to heaper til en. Resultatet vil da være den første heapen
        sammen med den andre heapen. Dette gjøres ved å kalle merge1 med
        rootnodene i begge heapene som parametere. Ikke noe forskjell her
        i forhold til MAW implementasjonen
        """
        if rhs == self:
            return
        self.root = self.merge1( self.root, rhs.root )
        rhs.root = None

        
    def merge1( self, node1, node2 ):
        """
        Merge1 sjekker om noen av nodene er null. Hvis det stemmer returnerer
        man den andre noden. Man kaller merge2 for å bygge heapen videre.
        merge2 blir kalt slik at parameter1 er den noden med lavest prioritet.
        Denne er også lik implementasjonen til MAW.
        """
        if node1 == None:
            return node2
        if node2 == None:
            return node1
        if node1.element < node2.element:
            return self.merge2(node1, node2)
        else:
            return self.merge2(node2, node1)

    def merge2( self, node1, node2 ):
        """
        Merge2 setter inn en node i treet. Hvis node1 ikke har noe venstre barn
        blir node2 satt til å være denne. Hvis node1 ikke har noe høyre barn 
        blir node2 satt til denne. Hvis node1 har både høyre og venstrebarn 
        merger vi node1 sitt venstre barn sammen med node2 hvis venstre barnet 
        har lik eller mindre npl verdi. Hvis ventrebarnet har større npl verdi 
        en høyre merger vi node1 sitt høyre  barn sammen med node 2. Oppdaterer
        samtidig node1 sin npl verdi til å være npl verdien til det motsatte 
        barnet pluss 1.
        """
        if node1.left is None:
            node1.left = node2
        elif node1.right is None:
            node1.right = node2
            node1.npl = 1
        else:
            if node1.left.npl <= node1.right.npl:
                node1.left = self.merge1(node1.left, node2)
                node1.npl = node1.right.npl + 1
            else:
                node1.right = self.merge1(node1.right, node2)
                node1.npl = node1.left.npl + 1
        return node1



if __name__ == '__main__':

    if len(sys.argv) != 3:
        print 'Not correct paramters given'
        sys.exit()

    # Åpner input filen
    try:
        infile = open(sys.argv[1],'r')
    except:
        print 'No shuch file:',sys.argv[1]

    # variabel for å holde heapen vi bygger.
    first = None

    # Lager en heap for hver linhe i file. Bruker
    # insert for å legge inn noder.
    for line in infile:

        # henter inn linjen fra file og gjør den om til
        # tallverdier.
        indata = [int(x) for x in line.split()]

        # Hvis linjen inneholder data lager vi en heap
        # med alle nummerene.
        if len(indata) != 0:
            # temporær lagring av heapen man bygger opp
            heap = LeftistHeap()
            for x in indata:
                heap.insert( x )

            # når vi har bygget opp heapen merger vi den vi
            # har laget med den first. Hvis first ikke eksistere
            # lar vi den temporære heapen være first.
            if first is None:
                first = heap
            else:
                first.merge( heap )

    data = None

    # Åpner output filen
    try:
        outfile = open(sys.argv[2], 'w')
    except:
        print 'Error on opening file:',sys.argv[2]
        sys.exit()

    outfile.write(' '.join(first.root.printTree())+'-'+' '.join(
            first.root.printNullPath())+'\n')

    
